/*
 * Minimal tls-library for dtbnc
 *
 */


#define DEFAULTCIPHERLIST       "ALL:!EXP"

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include "tls_dh.h"

char	*file_fullpath(char *fn);
char 	*tls_rsa_key_file = NULL;
char	*tls_rsa_cert_file = NULL;
char 	*tls_dsa_key_file = NULL;
char	*tls_dsa_cert_file = NULL;
char	*tls_dhparam_file = NULL;
char	*tls_rand_file = NULL;
char	*tls_cipher_list = DEFAULTCIPHERLIST;
SSL_CTX *ssl_ctx = NULL;
DH	*tmp_dh = NULL;
RSA	*tmp_rsa = NULL;


DH *tmp_dh_cb(SSL *ssl, int is_export, int keylength)
{
    FILE *fp;

    if (!tmp_dh) {
    	/* first try any 'tls_dhparam_file', else use built-in dh params */
    	if (tls_dhparam_file && (fp = fopen(tls_dhparam_file, "r"))) {
	    tmp_dh = PEM_read_DHparams(fp, NULL, NULL, NULL);
	    fclose(fp);
	    if (tmp_dh)
	    	return tmp_dh;
	}
	switch (keylength) {
	    case 512:	return tmp_dh = get_dh512();
	    case 768:	return tmp_dh = get_dh768();
	    case 1024:	return tmp_dh = get_dh1024();
	    case 1536:	return tmp_dh = get_dh1536();
	    case 2048:	return tmp_dh = get_dh2048();
	    default:	return tmp_dh = get_dh1024();
	}
    }
    else
    	return tmp_dh;
}


/* check_file() expands 'file' to an existing full path or NULL if not found */
void check_file(char **file)
{
    char *p;
    
    if (*file) {
    	p = file_fullpath(*file);
	if (p == *file)	/* same pointer returned from file_fullpath() */
	    return;
	free(*file);
	if (p) {
	    *file = malloc(strlen(p) + 1);
	    strcpy(*file, p);
	}
	else
	    *file = NULL;
    }
}



RSA *tmp_rsa_cb(SSL *ssl, int is_export, int keylength)
{
    if (!tmp_rsa)
	tmp_rsa = RSA_generate_key(keylength, RSA_F4, NULL, NULL);
    return tmp_rsa;
}





int seed_PRNG(void)
{
    char stackdata[1024];
    static char rand_file[200];
    FILE *fh;
    
#if OPENSSL_VERSION_NUMBER >= 0x00905100
    if (RAND_status())
	return 0;     /* PRNG already good seeded */
#endif
    /* if the device '/dev/urandom' is present, OpenSSL uses it by default.
     * check if it's present, else we have to make random data ourselfs.
     */
    if ((fh = fopen("/dev/urandom", "r"))) {
	fclose(fh);
	return 0;
    }
    /* the ftpd's rand file is (openssl-dir)/.rnd */
    snprintf(rand_file, sizeof(rand_file), "%s/.rnd", X509_get_default_cert_area());
    tls_rand_file = rand_file;
    if (!RAND_load_file(rand_file, 1024)) {
	/* no .rnd file found, create new seed */
	unsigned int c;
	c = time(NULL);
	RAND_seed(&c, sizeof(c));
	c = getpid();
	RAND_seed(&c, sizeof(c));
	RAND_seed(stackdata, sizeof(stackdata));
    }
#if OPENSSL_VERSION_NUMBER >= 0x00905100
    if (!RAND_status())
	return 2;   /* PRNG still badly seeded */
#endif
    return 0;
}







int tls_init(void)
{
    int	err;
	
#ifdef DEBUG_OPENSSL_MEM
    CRYPTO_malloc_debug_init();
    CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);
#endif /* DEBUG_OPENSSL_MEM */
    SSL_load_error_strings();
/* TODO: change to SSL_library_init() */
#if OPENSSL_VERSION_NUMBER < 0x00905100
    SSLeay_add_all_algorithms();
#else
    OpenSSL_add_all_algorithms();
#endif
    /* let's find out which files are available */
    check_file(&tls_rsa_cert_file);
    check_file(&tls_rsa_key_file);
    check_file(&tls_dsa_cert_file);
    check_file(&tls_dsa_key_file);
    check_file(&tls_dhparam_file);
    if (!tls_rsa_cert_file && !tls_dsa_cert_file) {
    	printf("No certificate files found!");
	return 2;
    }
    if (!tls_rsa_key_file)
    	tls_rsa_key_file = tls_rsa_cert_file;
    if (!tls_dsa_key_file)
    	tls_dsa_key_file = tls_dsa_cert_file;

    if (! tls_cipher_list)
	printf("WARNING: NULL tls_cipher_list!");
    if (seed_PRNG())
	printf("WARNING: Wasn't able to properly seed the PRNG!");

    ssl_ctx = SSL_CTX_new(SSLv23_method());
    if (!ssl_ctx) {
	printf("FATAL: SSL_CTX_new() %s",
		(char *)ERR_error_string(ERR_get_error(), NULL));
	return 1;
    }

    SSL_CTX_set_default_verify_paths(ssl_ctx);
    SSL_CTX_set_session_id_context(ssl_ctx, "1", 1);

    if (tls_rsa_cert_file) {
	err = SSL_CTX_use_certificate_file(ssl_ctx, tls_rsa_cert_file, X509_FILETYPE_PEM);
	if (err <= 0) {
	    printf("FATAL: SSL_CTX_use_certificate_file(%s) %s", tls_rsa_cert_file,
		(char *)ERR_error_string(ERR_get_error(), NULL));
	    return 3;
	}
	SSL_CTX_set_tmp_rsa_callback(ssl_ctx, tmp_rsa_cb);
    }
    if (tls_rsa_key_file) {
	err = SSL_CTX_use_PrivateKey_file(ssl_ctx, tls_rsa_key_file, X509_FILETYPE_PEM);
	if (err <= 0) {
	    printf("FATAL: SSL_CTX_use_PrivateKey_file(%s) %s", tls_rsa_key_file,
	    	(char *)ERR_error_string(ERR_get_error(), NULL));
	    return 4;
	}
    }
    if (tls_dsa_cert_file) {
	err = SSL_CTX_use_certificate_file(ssl_ctx, tls_dsa_cert_file, X509_FILETYPE_PEM);
	if (err <= 0) {
	    printf("FATAL: SSL_CTX_use_certificate_file(%s) %s", tls_dsa_cert_file,
		(char *)ERR_error_string(ERR_get_error(), NULL));
	    return 5;
	}
    }
    if (tls_dsa_key_file) {
	err = SSL_CTX_use_PrivateKey_file(ssl_ctx, tls_dsa_key_file, X509_FILETYPE_PEM);
	if (err <= 0) {
	    printf("FATAL: SSL_CTX_use_PrivateKey_file(%s) %s", tls_dsa_key_file,
	    	(char *)ERR_error_string(ERR_get_error(), NULL));
	    return 6;
	}
    }

    SSL_CTX_set_tmp_dh_callback(ssl_ctx, tmp_dh_cb);
    SSL_CTX_set_cipher_list(ssl_ctx, tls_cipher_list);

    return 0;
}



char *file_fullpath(char *fn)
{
    static char fp[256];
    FILE *file;
    char *dir;
    
    /* check if it is a full path already */
    if ((strchr(fn, '/'))) {
	if ((file = fopen(fn, "r"))) {
	    fclose(file);
	    return fn;
	}
	else
	    return NULL;
    }
    /* check if it is in current dir */
    if ((file = fopen(fn, "r"))) {
    	fclose(file);
	return fn;
    }
    if (!(dir = getenv(X509_get_default_cert_dir_env())))
    	dir = (char *)X509_get_default_cert_dir();
    snprintf(fp, sizeof(fp), "%s/%s", dir, fn);
    if ((file = fopen(fp, "r"))) {
    	fclose(file);
	return fp;
    }
    dir = (char *)X509_get_default_private_dir();
    snprintf(fp, sizeof(fp), "%s/%s", dir, fn);
    if ((file = fopen(fp, "r"))) {
    	fclose(file);
	return fp;
    }
    return NULL;
}







ssize_t tls_read(SSL *s, int fd, void *buf, size_t count) {
 ssize_t         cnt;

 if ( s != NULL ) {        
	if ((cnt = SSL_read(s, buf, count)) < 0 ) {
		switch ( SSL_get_error(s, cnt) ) {
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:               
				errno = EAGAIN;
				break;
			}
		cnt = -1;
		}
	return cnt;     
	} else {
	return read(fd, buf, count);
	}
}






ssize_t tls_write(SSL *s, int fd, void *buf, size_t count) {
 ssize_t         cnt;

 if ( s != NULL ) {
	if ((cnt = SSL_write(s, buf, count)) < 0 ) {
		switch ( SSL_get_error(s, cnt) ) {
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
				errno = EAGAIN;
				break;
			}
		cnt = -1;   
		}
	return cnt;      
	} else {                 
	return write(fd, buf, count);
	}                        
}
